/* ChordPro for Palm
 * (c) 2001 by Yih-Chun Hu
 *******************************************
 * TODO
 *    - local find
 *    - global find (do we want this?)
 *    - MemoPad import (is this a good feature?)
 *    - fix the slight chordwrap bug (when a full line w/o any chords is ouput
 *      in the "startwrap" area, then the chordwrap area kicks in, but the 
 *      chord wraps, wasting a line of space) It's an algorithm failure, though
 *      I've never seen it be a problem.
 *    - .pro / .crd conduit
 * CURRENT FEATURES
 *    - compression
 *    - beaming
 *    - transposition
 *
 *******************************************

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

 */

#define USE_ZLIB

#include <System/SystemPublic.h>
#include <UI/UIPublic.h>

#ifdef USE_ZLIB
#include <SysZLib.h>
#else
#include "PalmCompatibility.h"
#endif

#include "resource.h"
#include "common.h"

#define RX_BUF_SIZE		256
#define PREFERENCE_SAVED_Y	1
#define PREFERENCE_SAVED_SONG	2
#define PREFERENCE_SAVED_TRANS  3
#define SPACING			1
#define SCROLLY			100
#define SENSITIVITY		5
#define MINY			15
#define maxSongs		500
#define maxWidth		15
#define maxSongsxmaxWidth	7500
#define maxChordLength		50
#define MyWinDrawChars(s,c,x,y)	{if ((y)>=MINY) WinDrawChars(s,c,x,y);}
/*#define DebugMsg(s)		{WinDrawChars(s,StrLen(s),20,20);}*/
#define DebugMsg(s)		{1;}

static Int16 y;
static UInt32 pg;
static DmOpenRef ChordDB;
Boolean listNeedsUpdated = false;
void FixSongSelection(void);
static void InsertSongs(void);

static const char *song = "You need a database:[G]Got no [D]database "
			  "ton[C]ight[D]; [G]download a [D]generator "
			  "[C]to[D]day: http://www.yihchun.com/Palm/ChordPro/";

static UInt16 dbLocation[maxSongs];
static char   titles[maxSongsxmaxWidth];
static Int16  selectedSong = 0;
static Int16  currSongs = 0;

static char   transposed[maxChordLength+1];
static Int16  transposeAmount = 0;
static Boolean transposeSharp = true;
static Boolean disableTranspose = false;


const char  readChordLookup[7] = {0, 2, 3, 5, 7, 8, 10};

const char *readNote(const char *walk, const char *work, int *name) {
	*name = -1;

	if (walk == work)
		return walk;

	if (*walk >= 'A' && *walk <= 'G')
		*name = readChordLookup[*walk-'A'];
	else if (*walk >= 'a' && *walk <= 'g')
		*name = readChordLookup[*walk-'a'];

	if (*name >= 0) {
		walk++;
		if (walk != work && *walk == 'b') {
			walk++;
			(*name)--;
			if (*name == -1)
				*name = 11;
		}

		/* handle double flat */
		if (walk != work && *walk == 'b') {
			walk++;
			(*name)--;
			if (*name == -1)
				*name = 11;
		}

		if (walk != work && *walk == '#') {
			walk++;
			(*name)++;
			if (*name == 12)
				*name = 0;
		}

		if (walk != work && *walk == 'x') {
 			walk++;
			*name = (*name + 2) % 12;
		}

	}
	return walk;
}

void writeName(int name, int *i) {
	int j;

	for (j=1; j<7; j++)
		if (readChordLookup[j] > name)
			break;

	if (*i >= maxChordLength) return;

	if (transposeSharp || readChordLookup[j-1] == name)
		transposed[(*i)++] = 'A'+j-1;
	else
		transposed[(*i)++] = 'A'+(j%7);

	if (*i >= maxChordLength) return;

	if (readChordLookup[j-1] != name)
		transposed[(*i)++] = transposeSharp ? '#' : 'b';
}

const char *transpose(const char *crd, const char *work) {
	const char *walk = crd;
	int i = 0;
	int name = -1; /* A A# B C C# D D# E F F# G G# */

	if (disableTranspose) {
		while (walk != work && i < maxChordLength)
			transposed[i++] = *walk++;
		transposed[i++] = '\0';
		return transposed;
	}

	if (walk != work && *walk == '(')
		transposed[i++] = *walk++;

	walk = readNote(walk, work, &name);
	if (name != -1)
		writeName((name+transposeAmount)%12, &i);

	while (i < maxChordLength && walk != work && *walk != '/')
		transposed[i++] = *walk++;

	if (i == maxChordLength || walk == work) {
		transposed[i++] = 0;
		return transposed;
	}

	transposed[i++] = *walk++;
	walk = readNote(walk, work, &name);
	if (name != -1)
		writeName((name+transposeAmount)%12, &i);

	while (i < maxChordLength && walk != work)
		transposed[i++] = *walk++;

	transposed[i++] = 0;
	return transposed;
}

void * GetObjectPtr (UInt16 objectID) {
	FormPtr frm;
	frm = FrmGetActiveForm ();
	return (FrmGetObjectPtr (frm, FrmGetObjectIndex (frm, objectID)));
}

static Int16 DrawChordString( const char *str, Int16 y ) {
	Int16 baseliney = y; /* + FntBaseLine(); */
	Int16 baselyrline = baseliney + SPACING + FntCharHeight();
	Int16 x = 0;
	Int16 extentX, extentY;
	Int16 maxChars;
	Int16 chs = FntCharHeight() + SPACING;
	Int16 trLen;
	const char *work = str;
	const char *crd;
	const char *crdnew;
	
	WinGetDisplayExtent(&extentX, &extentY);

	if (baseliney > extentY)
		goto out_of_extent;

startwrap:
	if (!(crd = StrChr(work, '['))) {
		Int16 charsLeft = StrLen(work);
		while (charsLeft && (baseliney<extentY)) {
			maxChars = FntWordWrap (work, extentX-x);
			MyWinDrawChars(work, maxChars, x, baseliney);
			baseliney += chs;
			work += maxChars;
			while (*work == ' ') work++;
			charsLeft -= maxChars;
			x = 0;
		}
		return baseliney + chs;
	} else {
		while (crd != work) {
			maxChars = FntWordWrap (work, extentX-x);
			if (maxChars >= (crd-work)) {
				maxChars = crd-work;
				MyWinDrawChars(work, maxChars, x, baselyrline);
				x = FntCharsWidth(work, maxChars);
				break;
			} else {
				MyWinDrawChars(work, maxChars, x, baseliney);
				baseliney   += chs;
				baselyrline += chs;
				work += maxChars;
				while (*work == ' ') work++;
				if (baseliney > extentY)
					goto out_of_extent;
				x = 0;
			}
		}
	}

	while (*crd) {
		/*  cw             n
		 * [G]He is the Lor[C]d and He rei[G]gns on high
		 * crd points to the first letter in the chord
		 * work points to the first ]
		 * crdnew points to the next [
		 * our job is to line up 
		 */
		work = StrChr(++crd, ']');
		if (!work)
			break;

		crdnew = StrChr(work+1, '[');
		if (!crdnew)
			crdnew = work+StrLen(work);

		trLen = StrLen(transpose(crd, work));

		if (x && ((trLen > FntWordWrap(transposed, extentX-x)) ||
			  (*(work+1) && !FntWordWrap(work+1, extentX-x)))) {
			/* chord wrap */
			baseliney   = baselyrline + chs;
			baselyrline = baseliney   + chs;
			if (baseliney > extentY)
				goto out_of_extent;
			x = 0;
		}

		if (FntCharsWidth(transposed, trLen) > 
		    FntCharsWidth(work+1, crdnew-work-1)) {
			MyWinDrawChars(transposed, trLen, x, baseliney);
			MyWinDrawChars(work+1, crdnew-work-1, x, baselyrline);
			/* half space for chord heavy stuff */
			x += FntCharsWidth(transposed, trLen) + 
				FntCharWidth(' ')/2;
			if (x > extentX)
				x = extentX;
		} else {
			MyWinDrawChars(transposed, trLen, x, baseliney);
			maxChars = FntWordWrap(work+1, extentX-x);
			if (maxChars < crdnew-work-1) {
				MyWinDrawChars(work+1,maxChars,x,baselyrline);
				work = work + maxChars;
				while (*work == ' ') work++;
				x = 0;
				baseliney   = baselyrline + chs;
				baselyrline = baseliney   + chs;
				if (baseliney > extentY)
					goto out_of_extent;
				goto startwrap;
			}
			MyWinDrawChars(work+1, crdnew-work-1, x, baselyrline);
			x += FntCharsWidth(work+1, crdnew-work-1);
			if (x > extentX)
				x = extentX;
		}

		crd = crdnew;
	}
	return baselyrline + chs;

out_of_extent:
	WinDrawChars("*",1,extentX-FntCharWidth('*'),extentY-FntCharHeight());
	return baseliney;
}

#ifdef USE_ZLIB
voidpf zlib_alloc(voidpf opaque, uInt size, uInt num) {
	MemHandle  h;
	MemHandle *p;

	h = MemHandleNew(size*num+sizeof(MemHandle));
	if (!h) {
		DebugMsg("zalloc returning NULL; no handle...");
		return NULL;
	}

	p = MemHandleLock(h);
	if (!p) {
		DebugMsg("zalloc returning NULL; no lock...");
		MemHandleFree(h);
		return NULL;
	}

	*p = h;
	return p+1;
}

void zlib_free(voidpf opaque, voidpf address) {
	MemHandle *hp = ((MemHandle *)address)-1;
	MemHandle  h  = *hp;
	MemHandleUnlock(h);
	MemHandleFree(h);
}

void doUncompress(Ptr song, int songSize, Ptr target, int targetSize) {
	z_stream z;

	DebugMsg("Starting uncompress...");
	z.zalloc = zlib_alloc;
	z.zfree = zlib_free;
	z.opaque = NULL;
	z.total_in = 0;
	z.total_out = 0;
	DebugMsg("inits");
	inflateInit(&z);
	DebugMsg("inite");
	z.next_in = song;
	z.avail_in = songSize;
	z.next_out = target;
	z.avail_out = targetSize-1;
	DebugMsg("inflates");
	inflate(&z, Z_SYNC_FLUSH);
	DebugMsg("inflatee");

	/* let's participate in some termination... */
	((char *)target)[targetSize-1] = 0;
	DebugMsg("ends");
	inflateEnd(&z);
	DebugMsg("ende");
}

MemHandle uncompressSong(Ptr RecPointer) {
	MemHandle allocme;
	Ptr RealPointer;
	int targetSize = 1+*(Int16 *)RecPointer;

	/* allocate and nail down everything we need... */
	allocme = MemHandleNew(targetSize);
	if (!allocme) {
		WinDrawChars("We got more problems than you know.",
			     35, 20, 20);
		MemPtrUnlock(RecPointer);
		return NULL;
	}
	RealPointer = MemHandleLock(allocme);
	if (!RealPointer) {
		WinDrawChars("We got more problems than you know!",
			     35, 20, 20);
		MemHandleFree(allocme);
		MemPtrUnlock(RecPointer);
		return NULL;
	}

	/* gotta do some hardcore decompression... */
	doUncompress(((Int16 *)RecPointer)+1, 
		     MemPtrSize(RecPointer)-sizeof(Int16),
		     RealPointer, targetSize);

	MemHandleUnlock(allocme);
	return allocme;
}
#endif

static void DisplaySong() {
	VoidHand RecHandle;
	Ptr RecPointer;
	MemHandle allocme;
	Ptr RealPointer; /* not a pointer to Real... */
	char *colon;
	Int16 retval;

	WinEraseWindow();
	FrmDrawForm(FrmGetActiveForm());

	RecHandle = DmQueryRecord (ChordDB, dbLocation[selectedSong]);
	RecPointer = MemHandleLock(RecHandle);
	if (*(Int16 *)RecPointer) {
#ifdef USE_ZLIB
		allocme = uncompressSong(RecPointer);
		if (!allocme) {
			DmReleaseRecord(ChordDB, dbLocation[selectedSong], 
					false);
			return;
		}

		/* this better work... */
		RealPointer = MemHandleLock(allocme);
		if (!RealPointer) {
			DmReleaseRecord(ChordDB, dbLocation[selectedSong], 
					false);
			MemHandleFree(allocme);
			return;
		}

		colon = StrChr(RealPointer, ':');
		retval = DrawChordString(colon ? (colon+1) : RealPointer,y);

		/* Save the whales, free the mallocs... */
		MemHandleUnlock(allocme);
		MemHandleFree(allocme);
#else
		WinDrawChars("You need zlib!", 14, 20, 20);
#endif
	} else {
		RealPointer = ((Int16 *)RecPointer)+1;
		colon = StrChr(RealPointer, ':');
		retval = DrawChordString(colon ? (colon+1) : RealPointer,y);
	}
	MemPtrUnlock(RecPointer);
	DmReleaseRecord(ChordDB, dbLocation[selectedSong], false);

	if (retval < MINY) {
		if (currSongs > (selectedSong + 1)) {
			selectedSong++;
			LstSetSelection(GetObjectPtr(songList), selectedSong);
			y = MINY;
			FixSongSelection();
		} else {
			y = MINY;
		}
	} else if (retval == y) {
		if (selectedSong) {
			selectedSong--;
			LstSetSelection(GetObjectPtr(songList), selectedSong);
			y = MINY;
			FixSongSelection();
		} else {
			y = MINY;
		}
	}
}

void FixSongSelection() {
	ListType *list = GetObjectPtr(songList);
	char label[maxWidth];
	Int16 pixwidth, strlength;
	Boolean trash;

	selectedSong = LstGetSelection(list);
	LstSetTopItem(list, selectedSong);
	StrNCopy(label, &titles[selectedSong * maxWidth], maxWidth);
	label[maxWidth] = 0;
	strlength = StrLen(label);
	pixwidth = SONGTITLEWIDTH;
	FntCharsInWidth (label, &pixwidth, &strlength, &trash);
	label[strlength] = 0;
	CtlSetLabel(GetObjectPtr(songListTrigger), label);
	DisplaySong();
}

static void songListDrawFunc(UInt itemNum, RectanglePtr bounds, CharPtr *data){
	WinDrawChars(&titles[itemNum * maxWidth], 
		     StrLen(&titles[itemNum * maxWidth]),
		     bounds->topLeft.x, bounds->topLeft.y);
}

static void UpdateList(Boolean justOpened) {
	char    *ptitle[maxSongs];
	UInt16	 songMax;
	UInt16   song;
	Int16	 copychars;
	Int16	 sz;
	UInt16   attribs;
	UInt16	 i,j;
	UInt16   jxmw;
	VoidHand RecHandle;
	Int16   *ORecPointer;
	char	*RecPointer;
	char	*colonPointer;
	ListType*list;
	char     waitmsg[34];

top:
	songMax = DmNumRecords(ChordDB);
	song = j = jxmw = 0;
	while (song < songMax && j != maxSongs) {

		if (!(song & 7)) {
			StrPrintF(waitmsg, "Getting title from song %4d/%4d", 
				  song, songMax);
			WinEraseWindow();
			WinDrawChars(waitmsg, StrLen(waitmsg), 20, 100);
		}

		DmRecordInfo(ChordDB, song, &attribs, NULL, NULL);
		if (attribs & dmRecAttrDelete)
			continue;
		dbLocation[j] = song;
		RecHandle = DmQueryRecord(ChordDB, song);
		ORecPointer = MemHandleLock(RecHandle);
		if (*ORecPointer) {
#ifdef USE_ZLIB
			doUncompress(ORecPointer+1, 
				     MemPtrSize(ORecPointer)-sizeof(Int16),
				     titles+jxmw, maxWidth-1);
			for (i=0; i<maxWidth; i++)
				if (titles[i+jxmw] == ':') {
					titles[i+jxmw] = 0;
					break;
				}
#endif
		} else {
			RecPointer = (char *)(ORecPointer+1);
			colonPointer = StrChr(RecPointer, ':');
			if (!colonPointer)
				colonPointer = RecPointer + StrLen(RecPointer);
			copychars = colonPointer - RecPointer;
			if (copychars >= maxWidth)
				copychars = maxWidth-1;
			for (i=0; i<copychars; i++)
				titles[jxmw + i] = RecPointer[i];
			titles[jxmw+copychars] = 0;
		}
		MemPtrUnlock(ORecPointer);
		DmReleaseRecord(ChordDB, song, false);
		j++; jxmw += maxWidth;
		song++;
	}
	for (i=0;i<j;i++)
		ptitle[i] = &titles[i*maxWidth];

	currSongs = j;

	if (!currSongs) {
		InsertSongs();
		/* this was tail recursive: UpdateList(justOpened); return; */
		goto top;
	}

	/* if justOpened, try opening it as a preference */
	sz = sizeof(selectedSong);
	if (!justOpened ||
	    PrefGetAppPreferences(appCreatorID, PREFERENCE_SAVED_SONG,
				  &selectedSong, &sz, 
				  true) == noPreferenceFound ||
	    selectedSong < 0 ||
	    selectedSong >= currSongs)
		selectedSong = 0;

	list = GetObjectPtr(songList);
	LstSetListChoices(list, ptitle, j);
	LstSetSelection(list, selectedSong);
	LstSetDrawFunction(list, songListDrawFunc);
	FixSongSelection();
}

static Boolean aboutHandler(EventType *event) {
	switch (event->eType) {
	case frmOpenEvent:
	case frmUpdateEvent:
		FrmDrawForm(FrmGetActiveForm());
		return true;
		break;

	case ctlSelectEvent:
		if (event->data.ctlSelect.controlID == infoFormOkButton) {
			FrmReturnToForm(mainForm);
			return true;
		}
		break;
	default: break;
	}
	return false;
}

static void BeamCurrentSong() {
	Err err;
	ExgSocketType theSocket;
	VoidHand RecHandle;
	Ptr RecPointer;
	char title[maxWidth+1];

	RecHandle = DmQueryRecord(ChordDB, dbLocation[selectedSong]);
	RecPointer = MemHandleLock(RecHandle);

	StrNCopy(title, &titles[maxWidth*selectedSong], maxWidth);
	title[maxWidth] = 0;

	MemSet(&theSocket, sizeof(theSocket), 0);
	theSocket.type = "application/x-chordpro";
	theSocket.description = title;

	err = ExgPut(&theSocket);
	if (!err)
		ExgSend(&theSocket, RecPointer, MemPtrSize(RecPointer), &err);
                                                
	err = ExgDisconnect(&theSocket, err);

	MemHandleUnlock(RecHandle);
	DmReleaseRecord(ChordDB, dbLocation[selectedSong], false);
}

static void ReceiveBeam(ExgSocketPtr socketPtr, int subcall) {
	Err err;
	DmOpenRef BeamChordDB;
	ULong bytesRxd=0;
	Int16 totalBytes;
	VoidHand RecHandle;
	Ptr RecPointer;
	UInt16 index = 65535;
	char buf[RX_BUF_SIZE];
	int offset = 0;

	if (!subcall)
		BeamChordDB = DmOpenDatabaseByTypeCreator(dbType, appCreatorID,
							  dmModeReadWrite);
	else
		BeamChordDB = ChordDB;

	if (!BeamChordDB)
		return;

	if (ExgAccept(socketPtr))
		goto error;

	if (sizeof(totalBytes) != ExgReceive(socketPtr, &totalBytes,
					     sizeof(totalBytes), &err))
		goto error;

#ifndef USE_ZLIB
	if (totalBytes)
		goto error;
#endif

	RecHandle = DmNewRecord(BeamChordDB, &index, sizeof(totalBytes));
	RecPointer = MemHandleLock(RecHandle);
	DmWrite(RecPointer, 0, &totalBytes, sizeof(totalBytes));
	MemPtrUnlock(RecPointer);

	do {
		bytesRxd = ExgReceive(socketPtr, buf, RX_BUF_SIZE, &err);
		RecHandle = DmResizeRecord(BeamChordDB, index, 
					   offset + 2 + bytesRxd);
		RecPointer = MemHandleLock(RecHandle);
		DmWrite(RecPointer, offset + 2, buf, bytesRxd);
		MemPtrUnlock(RecPointer);
		offset += bytesRxd;
	} while (bytesRxd);

	DmReleaseRecord(BeamChordDB, index, true);	

	if (subcall)
		listNeedsUpdated = true;

error:
	if (!subcall)
		DmCloseDatabase(BeamChordDB);
}

static Boolean mainHandler(EventType *event) {
	switch (event->eType) {
	case frmOpenEvent:
		FrmDrawForm(FrmGetActiveForm());
		DebugMsg("UpdateList: frmOpenEvent");
		UpdateList(true);
		FixSongSelection(); /* implicitly DisplaySong()s */
		return true;
		break;

	case frmUpdateEvent:
		if (listNeedsUpdated) {
			listNeedsUpdated = false;
			DebugMsg("UpdateList: frmUpdateEvent");
			UpdateList(false);
		}
		FixSongSelection(); /* implicitly DisplaySong()s */
		return true;
		break;

	case ctlSelectEvent:
		if (event->data.ctlSelect.controlID == songListTrigger) {
			LstPopupList(GetObjectPtr(songList));
			y = MINY;
			FixSongSelection();
			return true;
		}
		break;

	case keyDownEvent:
		if (TimGetTicks() > (pg+SysTicksPerSecond()/SENSITIVITY)) {
			if (event->data.keyDown.chr == vchrPageUp)
				y += SCROLLY;
			if (event->data.keyDown.chr == vchrPageDown)
				y -= SCROLLY;
			if (event->data.keyDown.chr == vchrPageUp ||
			    event->data.keyDown.chr == vchrPageDown) {
				pg = TimGetTicks();
				FixSongSelection();
			}

			if ((event->data.keyDown.chr >= 'a' &&
			     event->data.keyDown.chr <= 'z') ||
			    (event->data.keyDown.chr >= 'A' &&
			     event->data.keyDown.chr <= 'Z')) {
				Int16 i;
				char c = event->data.keyDown.chr;
				for (i=0;i<currSongs;i++)
					if (titles[i*maxWidth] == c)
						break;

				if (i != currSongs) {
					selectedSong = i;
					LstSetSelection(GetObjectPtr(songList),
							selectedSong);
					FixSongSelection();
				}
			}
		}
		return true;
		break;

	case menuEvent:
		switch (event->data.menu.itemID) {
		case menuItemAbout:
			FrmPopupForm(aboutForm); 
			return true; break;
		case menuItemDelete:
			DmRemoveRecord(ChordDB, dbLocation[selectedSong]);
			DebugMsg("UpdateList: menuItemDelete");
			UpdateList(false);
			return true; break;
		case menuTransposeUp:
			transposeAmount++; 
			FixSongSelection(); return true; break;
		case menuStopTranspose:
			transposeAmount = 0; 
			FixSongSelection(); return true; break;
		case menuTransposeDown:
			transposeAmount = (transposeAmount+11)%12; 
			FixSongSelection(); return true; break;
		case menuSharpFlat:
			transposeSharp = !transposeSharp;
			FixSongSelection(); return true; break;
		case menuDisableTranspose:
			disableTranspose = !disableTranspose;
			FixSongSelection(); return true; break;
		case menuBeamSong:
			BeamCurrentSong();
			return true; break;
		}
	default: break;
	}

	return false;
}

static void EventLoop() {
	short	  err;
	int	  formID;
	FormPtr   form;
	EventType event; 

	do {
		EvtGetEvent( &event, evtWaitForever );

		if (SysHandleEvent(&event)) continue;
		if (MenuHandleEvent((void *)0, &event, &err)) continue;

		if (event.eType == frmLoadEvent) {
			formID = event.data.frmLoad.formID;
			form = FrmInitForm(formID);
			FrmSetActiveForm(form);

			switch(formID) {
			case mainForm: 
				FrmSetEventHandler(form, mainHandler); break;
			case aboutForm:	
				FrmSetEventHandler(form, aboutHandler); break;
			}
		}

		FrmDispatchEvent(&event);

	} while (event.eType != appStopEvent);
	return;
}

static void StopApplication() {
	FrmCloseAllForms();
	DmCloseDatabase(ChordDB);
	PrefSetAppPreferences(appCreatorID, PREFERENCE_SAVED_Y, 0, 
				&y, sizeof(y), true);
	PrefSetAppPreferences(appCreatorID, PREFERENCE_SAVED_SONG, 0,
				&selectedSong, sizeof(selectedSong), true);
	PrefSetAppPreferences(appCreatorID, PREFERENCE_SAVED_TRANS, 0,
				&transposeAmount, sizeof(transposeAmount), 
				true);
}

static void InsertSongs() {
	UInt index = 0;
	VoidHand RecHandle;
	Ptr RecPointer;
	const Int16 zero = 0;

	RecHandle = DmNewRecord(ChordDB, &index, StrLen(song)+3);
	RecPointer = MemHandleLock(RecHandle);
	DmWrite(RecPointer, 0, &zero, sizeof(zero));
	DmWrite(RecPointer, 2, song, StrLen(song)+1);
	MemPtrUnlock(RecPointer);
	DmReleaseRecord(ChordDB, index, true);
}

static void StartApplication() {
	Int16 sz;

	/* since the emulator doesn't hotsync... */
	ExgRegisterData(appCreatorID, exgRegTypeID, "application/x-chordpro");

	FrmGotoForm(mainForm);

#ifdef USE_ZLIB
	ZLSetup;
#endif

	sz = sizeof(y);
	if (PrefGetAppPreferences(appCreatorID, PREFERENCE_SAVED_Y,
				  &y, &sz, true) == noPreferenceFound)
		y = MINY;

	if (PrefGetAppPreferences(appCreatorID, PREFERENCE_SAVED_TRANS,
				  &transposeAmount, &sz, true) == 
	    noPreferenceFound)
		transposeAmount = 0;

	ChordDB = DmOpenDatabaseByTypeCreator(dbType, appCreatorID,
						dmModeReadWrite);
	if (!ChordDB) {
		DmCreateDatabase(0, dbName, appCreatorID, dbType, false);
		ChordDB = DmOpenDatabaseByTypeCreator(dbType, appCreatorID,
							dmModeReadWrite);
		InsertSongs();
	}
}

DWord PilotMain (Word cmd, Ptr cmdPBP, Word launchFlags) {
	if (cmd == sysAppLaunchCmdNormalLaunch) {
		StartApplication();
		EventLoop();
		StopApplication ();
		ExgRegisterData(appCreatorID, exgRegTypeID, 
				"application/x-chordpro");
	} else if (cmd == sysAppLaunchCmdSyncNotify) {
		ExgRegisterData(appCreatorID, exgRegTypeID, 
				"application/x-chordpro");
	} else if (cmd == sysAppLaunchCmdExgReceiveData) {
		ReceiveBeam(cmdPBP, launchFlags & sysAppLaunchFlagSubCall);
	}

	return 0;
}
